//
//  metafile.cpp
//
//  Classes to create Windows metafiles from non-Windows programs
//
//  Copyright (c) 1995 by JERICO Systems - All rights reserved
//

#include <stdlib.h>
#include <stdio.h>
#include <mem.h>
#include <string.h>

#include "metafile.h"

#ifndef __WINDOWS_H

// Set Graphics Parameters
#define META_SETBKCOLOR              0x0201
#define META_SETBKMODE               0x0102
#define META_SETMAPPERFLAGS          0x0231
#define META_SETMAPMODE              0x0103
#define META_SETPOLYFILLMODE         0x0106
#define META_SETROP2                 0x0104
#define META_SETRELABS               0x0105
#define META_SETSTRETCHBLTMODE       0x0107
#define META_SETTEXTALIGN            0x012E
#define META_SETTEXTCHAREXTRA        0x0108
#define META_SETTEXTCOLOR            0x0209
#define META_SETTEXTJUSTIFICATION    0x020A
#define META_SETVIEWPORTORG          0x020D
#define META_SETVIEWPORTEXT          0x020E
#define META_SETWINDOWORG            0x020B
#define META_SETWINDOWEXT            0x020C

#define META_OFFSETVIEWPORTORG       0x0211
#define META_OFFSETWINDOWORG         0x020F
#define META_SCALEVIEWPORTEXT        0x0412
#define META_SCALEWINDOWEXT          0x0410

#define META_EXCLUDECLIPRECT         0x0415
#define META_INTERSECTCLIPRECT       0x0416

// Drawing/Filling Functions
#define META_ARC                     0x0817
#define META_CHORD                   0x0830
#define META_DRAWTEXT                0x062F
#define META_ELLIPSE                 0x0418
#define META_EXTFLOODFILL            0x0548
#define META_EXTTEXTOUT              0x0a32
#define META_FLOODFILL               0x0419
#define META_LINETO                  0x0213
#define META_MOVETO                  0x0214
#define META_PIE                     0x081A
#define META_POLYGON                 0x0324
#define META_POLYLINE                0x0325
#define META_POLYPOLYGON             0x0538
#define META_RECTANGLE               0x041B
#define META_ROUNDRECT               0x061C
#define META_SETPIXEL                0x041F
#define META_TEXTOUT                 0x0521

// Bitblit Functions
#define META_BITBLT                  0x0922
#define META_PATBLT                  0x061D
#define META_STRETCHBLT              0x0B23
#define META_DIBBITBLT               0x0940
#define META_DIBSTRETCHBLT           0x0b41

// Bitmap Functions
#define META_SETDIBTODEV             0x0d33
#define META_STRETCHDIB              0x0f43

// Region Functions
#define META_FILLREGION              0x0228
#define META_FRAMEREGION             0x0429
#define META_INVERTREGION            0x012A
#define META_PAINTREGION             0x012B

#define META_OFFSETCLIPRGN           0x0220
#define META_SELECTCLIPREGION        0x012C

// Device Context Functions
#define META_SAVEDC                  0x001E
#define META_RESTOREDC               0x0127
#define META_RESETDC                 0x014C

// Palette Functions
#define META_SELECTPALETTE           0x0234
#define META_REALIZEPALETTE          0x0035
#define META_ANIMATEPALETTE          0x0436
#define META_SETPALENTRIES           0x0037
#define META_RESIZEPALETTE           0x0139

// Printer Control Functions
#define META_ABORTDOC                0x0052
#define META_STARTDOC                0x014D
#define META_STARTPAGE               0x004F
#define META_ENDPAGE                 0x0050
#define META_ENDDOC                  0x005E

// Create/Delete/Select Graphic Drawing Objects
#define META_CREATEPALETTE           0x00f7
#define META_CREATEBRUSH             0x00F8
#define META_CREATEPATTERNBRUSH      0x01F9
#define META_CREATEPENINDIRECT       0x02FA
#define META_CREATEFONTINDIRECT      0x02FB
#define META_CREATEBRUSHINDIRECT     0x02FC
#define META_CREATEBITMAPINDIRECT    0x02FD
#define META_CREATEBITMAP            0x06FE
#define META_CREATEREGION            0x06FF
#define META_DIBCREATEPATTERNBRUSH   0x0142
#define META_DELETEOBJECT            0x01f0
#define META_SELECTOBJECT            0x012D

#define META_ESCAPE                  0x0626

#endif
/////////////////////////////////////////////////////////////////////////////
//                           WinMetaFile Class

WinMetaFile::WinMetaFile( char* fileName, WORD Inch ) : nextIndex(0)
{
   metaFile = fopen( fileName, "w+b" );

   preHeader.key = 0x9AC6CDD7;
   preHeader.unused = 0;
   preHeader.box.x1 = 0;
   preHeader.box.y1 = 0;
   preHeader.box.x2 = 0;
   preHeader.box.y2 = 0;
   preHeader.inch = Inch;
   preHeader.reserved = 0;
   preHeader.checkSum = 0;

   fwrite( &preHeader, sizeof preHeader, 1, metaFile );

   header.type = 1;
   header.headerSize = sizeof header / sizeof( WORD );
   header.version = 0x300;
   header.fileSize = sizeof header / sizeof( WORD );
   header.maxObjectCount = 0;
   header.maxRecord = 0;
   header.unused = 0;

   fwrite( &header, sizeof header, 1, metaFile );
}


WinMetaFile::~WinMetaFile()
{
   DoStdRecord( 3, 0 );

   WORD * w = (WORD*)&preHeader;
   do { preHeader.checkSum ^= *w++; } while ( w != &preHeader.checkSum );

   fseek( metaFile, 0, 0 );
   fwrite( &preHeader, sizeof preHeader, 1, metaFile );
   fwrite( &header, sizeof header, 1, metaFile );
   fclose( metaFile );
}


void WinMetaFile::UpdatePreHeader( int x1, int y1, int x2, int y2 )
{
   if ( x1 < preHeader.box.x1 ) preHeader.box.x1 = x1;
   if ( y1 < preHeader.box.y1 ) preHeader.box.y1 = y1;
   if ( x2 > preHeader.box.x2 ) preHeader.box.x2 = x2;
   if ( y2 > preHeader.box.y2 ) preHeader.box.y2 = y2;
}


void WinMetaFile::UpdateHeader( DWORD recSize, WORD newObjects )
{
   header.fileSize += recSize;
   header.maxObjectCount += newObjects;
   if ( recSize > header.maxRecord ) header.maxRecord = recSize;
}


void WinMetaFile::DoStdRecord( DWORD size, WORD code, WORD newObjects )
{
   UpdateHeader( size, newObjects );
   fwrite( &size, sizeof size, 1, metaFile );
   fwrite( &code, sizeof code, 1, metaFile );
}


WORD WinMetaFile::CreateBrushIndirect( LOGBRUSH & brush )
{
   DoStdRecord( 3 + (sizeof brush + 1) / sizeof(WORD),
                META_CREATEBRUSHINDIRECT, 1 );
   fwrite( &brush, sizeof brush, 1, metaFile );
   return nextIndex++;
}


WORD WinMetaFile::CreatePenIndirect( LOGPEN & pen )
{
   DoStdRecord( 3 + (sizeof pen + 1) / sizeof(WORD),
                META_CREATEPENINDIRECT, 1 );
   fwrite( &pen, sizeof pen, 1, metaFile );
   return nextIndex++;
}


WORD WinMetaFile::CreateFontIndirect( LOGFONT & font )
{
   DoStdRecord( 3 + (sizeof font + 1) / sizeof(WORD),
                META_CREATEFONTINDIRECT, 1 );
   fwrite( &font, sizeof font, 1, metaFile );
   return nextIndex++;
}


WORD WinMetaFile::CreateSolidBrush( COLORREF color )
{
   DoStdRecord( 7, META_CREATEBRUSHINDIRECT, 1 );
   WORD  style = 0, hatch = 0;
   fwrite( &style, sizeof style, 1, metaFile );
   fwrite( &color, sizeof color, 1, metaFile );
   fwrite( &hatch, sizeof hatch, 1, metaFile );
   return nextIndex++;
}


WORD WinMetaFile::CreatePen( WORD style, POINT width, COLORREF color )
{
   DoStdRecord( 8, META_CREATEPENINDIRECT, 1 );
   fwrite( &style, sizeof style, 1, metaFile );
   fwrite( &width, sizeof width, 1, metaFile );
   fwrite( &color, sizeof color, 1, metaFile );
   return nextIndex++;
}


void WinMetaFile::Arc( int x1, int y1, int x2, int y2,
                       int xbeg, int ybeg, int xend, int yend )
{
   DoStdRecord( 11, META_ARC );
   fwrite( &yend, sizeof yend, 1, metaFile );
   fwrite( &xend, sizeof xend, 1, metaFile );
   fwrite( &ybeg, sizeof ybeg, 1, metaFile );
   fwrite( &xbeg, sizeof xbeg, 1, metaFile );
   fwrite( &y2, sizeof y2, 1, metaFile );
   fwrite( &x2, sizeof x2, 1, metaFile );
   fwrite( &y1, sizeof y1, 1, metaFile );
   fwrite( &x1, sizeof x1, 1, metaFile );
   UpdatePreHeader( x1, y1, x2, y2 );    // NOT ACCURATE, must improve
}


void WinMetaFile::Chord( int x1, int y1, int x2, int y2,
                         int xbeg, int ybeg, int xend, int yend )
{
   DoStdRecord( 11, META_CHORD );
   fwrite( &yend, sizeof yend, 1, metaFile );
   fwrite( &xend, sizeof xend, 1, metaFile );
   fwrite( &ybeg, sizeof ybeg, 1, metaFile );
   fwrite( &xbeg, sizeof xbeg, 1, metaFile );
   fwrite( &y2, sizeof y2, 1, metaFile );
   fwrite( &x2, sizeof x2, 1, metaFile );
   fwrite( &y1, sizeof y1, 1, metaFile );
   fwrite( &x1, sizeof x1, 1, metaFile );
   UpdatePreHeader( x1, y1, x2, y2 );    // NOT ACCURATE, must improve
}


void WinMetaFile::Pie( int x1, int y1, int x2, int y2,
                       int xbeg, int ybeg, int xend, int yend )
{
   DoStdRecord( 11, META_PIE );
   fwrite( &yend, sizeof yend, 1, metaFile );
   fwrite( &xend, sizeof xend, 1, metaFile );
   fwrite( &ybeg, sizeof ybeg, 1, metaFile );
   fwrite( &xbeg, sizeof xbeg, 1, metaFile );
   fwrite( &y2, sizeof y2, 1, metaFile );
   fwrite( &x2, sizeof x2, 1, metaFile );
   fwrite( &y1, sizeof y1, 1, metaFile );
   fwrite( &x1, sizeof x1, 1, metaFile );
   UpdatePreHeader( x1, y1, x2, y2 );    // NOT ACCURATE, must improve
}


void WinMetaFile::Ellipse( int x1, int y1, int x2, int y2 )
{
   DoStdRecord( 7, META_ELLIPSE );
   fwrite( &y2, sizeof y2, 1, metaFile );
   fwrite( &x2, sizeof x2, 1, metaFile );
   fwrite( &y1, sizeof y1, 1, metaFile );
   fwrite( &x1, sizeof x1, 1, metaFile );
   UpdatePreHeader( x1, y1, x2, y2 );
}


void WinMetaFile::ExtFloodFill( int x, int y, COLORREF color, UINT type )
{
   DoStdRecord( 8, META_EXTFLOODFILL );
   fwrite( &type, sizeof type, 1, metaFile );
   fwrite( &color, sizeof color, 1, metaFile );
   fwrite( &y, sizeof y, 1, metaFile );
   fwrite( &x, sizeof x, 1, metaFile );
}


void WinMetaFile::ExtTextOut( int x, int y, UINT option, RECT & rect,
                              char * text, UINT count, int * Dx )
{

   DoStdRecord( 7 + (count+1)/2 + (option==0 ? 0 : 4) +
                (Dx==NULL ? 0 : count), META_EXTTEXTOUT );
   fwrite( &y, sizeof y, 1, metaFile );
   fwrite( &x, sizeof x, 1, metaFile );
   fwrite( &count, sizeof count, 1, metaFile );
   fwrite( &option, sizeof option, 1, metaFile );
   if ( option != 0 )
      fwrite( &rect, sizeof rect, 1, metaFile );
   fwrite( text, count, 1, metaFile );
   if ( count % 2 != 0 )
        fwrite( " ", 1, 1, metaFile );
   if ( Dx != NULL )
        fwrite( (int*)Dx, count * sizeof(int), 1, metaFile );
   UpdatePreHeader( x, y, x, y );         // NOT CORRECT, must improve
}


void WinMetaFile::FloodFill( int x, int y, COLORREF color )
{
   DoStdRecord( 7, META_FLOODFILL );
   fwrite( &color, sizeof color, 1, metaFile );
   fwrite( &y, sizeof y, 1, metaFile );
   fwrite( &x, sizeof x, 1, metaFile );
}


void WinMetaFile::LineTo( int x, int y )
{
   DoStdRecord( 5, META_LINETO );
   fwrite( &y, sizeof y, 1, metaFile );
   fwrite( &x, sizeof x, 1, metaFile );
   UpdatePreHeader( x, y, x, y );
}


void WinMetaFile::MoveTo( int x, int y )
{
   DoStdRecord( 5, META_MOVETO );
   fwrite( &y, sizeof y, 1, metaFile );
   fwrite( &x, sizeof x, 1, metaFile );
   UpdatePreHeader( x, y, x, y );
}


void WinMetaFile::Polygon( const POINT * pt, int count )
{
   DoStdRecord( 4 + 2*count, META_POLYGON );
   fwrite( &count, sizeof count, 1, metaFile );
   fwrite( (POINT*)pt, count * sizeof(POINT), 1, metaFile );
   for ( int i = 0;  i < count;  i++ )
       UpdatePreHeader( pt[i].x, pt[i].y, pt[i].x, pt[i].y );
}


void WinMetaFile::Polyline( const POINT * pt, int count )
{
   DoStdRecord( 4 + 2*count, META_POLYLINE );
   fwrite( &count, sizeof count, 1, metaFile );
   fwrite( (POINT*)pt, count * sizeof(POINT), 1, metaFile );
   for ( int i = 0;  i < count;  i++ )
       UpdatePreHeader( pt[i].x, pt[i].y, pt[i].x, pt[i].y );
}


void WinMetaFile::PolyPolygon( const POINT * pt, int * counts, int count )
{
   int i, nPts = 0;
   for ( i = 0;  i < count;  i++ )
       nPts += counts[i];
   DoStdRecord( 4 + count + 2*nPts, META_POLYPOLYGON );
   fwrite( &count, sizeof count, 1, metaFile );              // # polygons
   fwrite( (int*)counts, count * sizeof(int), 1, metaFile ); // pt counts
   fwrite( (POINT*)pt, nPts * sizeof(POINT), 1, metaFile );  // all pts
   for ( i = 0;  i < nPts;  i++ )
       UpdatePreHeader( pt[i].x, pt[i].y, pt[i].x, pt[i].y );
}


void WinMetaFile::PolyPolyline( POINT ** curves, uint * counts, uint count )
{
   for ( int c = 0;  c < count;  c++ )
   {
       Polyline( curves[c], counts[c] );
   }
}


void WinMetaFile::Rectangle( int x1, int y1, int x2, int y2 )
{
   DoStdRecord( 7, META_RECTANGLE );
   fwrite( &y2, sizeof y2, 1, metaFile );
   fwrite( &x2, sizeof x2, 1, metaFile );
   fwrite( &y1, sizeof y1, 1, metaFile );
   fwrite( &x1, sizeof x1, 1, metaFile );
   UpdatePreHeader( x1, y1, x2, y2 );
}


void WinMetaFile::RoundRect( int x1, int y1, int x2, int y2, int eWide, int eHigh )
{
   DoStdRecord( 9, META_ROUNDRECT );
   fwrite( &eHigh, sizeof eHigh, 1, metaFile );
   fwrite( &eWide, sizeof eWide, 1, metaFile );
   fwrite( &y2, sizeof y2, 1, metaFile );
   fwrite( &x2, sizeof x2, 1, metaFile );
   fwrite( &y1, sizeof y1, 1, metaFile );
   fwrite( &x1, sizeof x1, 1, metaFile );
   UpdatePreHeader( x1, y1, x2, y2 );
}


void WinMetaFile::SelectObject( WORD index )
{
   DoStdRecord( 4, META_SELECTOBJECT );
   fwrite( &index, sizeof index, 1, metaFile );
}


void WinMetaFile::SetBkColor( COLORREF color )
{
   DoStdRecord( 5, META_SETBKCOLOR );
   fwrite( &color, sizeof color, 1, metaFile );
}


void WinMetaFile::SetBkMode( int mode )
{
   DoStdRecord( 4, META_SETBKMODE );
   fwrite( &mode, sizeof mode, 1, metaFile );
}


void WinMetaFile::SetMapMode( WORD mode )
{
   DoStdRecord( 4, META_SETMAPMODE );
   fwrite( &mode, sizeof mode, 1, metaFile );
}


void WinMetaFile::SetROP2( int mode )
{
   DoStdRecord( 4, META_SETROP2 );
   fwrite( &mode, sizeof mode, 1, metaFile );
}


void WinMetaFile::SetWindowExt( int x, int y )
{
   DoStdRecord( 5, META_SETWINDOWEXT );
   fwrite( &y, sizeof y, 1, metaFile );
   fwrite( &x, sizeof x, 1, metaFile );
}


void WinMetaFile::SetWindowOrg( int x, int y )
{
   DoStdRecord( 5, META_SETWINDOWORG );
   fwrite( &y, sizeof y, 1, metaFile );
   fwrite( &x, sizeof x, 1, metaFile );
}


void WinMetaFile::SetTextAlign( UINT align )
{
   DoStdRecord( 4, META_SETTEXTALIGN );
   fwrite( &align, sizeof align, 1, metaFile );
}


void WinMetaFile::SetTextColor( COLORREF color )
{
   DoStdRecord( 5, META_SETTEXTCOLOR );
   fwrite( &color, sizeof color, 1, metaFile );
}


void WinMetaFile::SetTextJustification( int space, int breaks )
{
   DoStdRecord( 5, META_SETTEXTJUSTIFICATION );
   fwrite( &breaks, sizeof breaks, 1, metaFile );
   fwrite( &space, sizeof space, 1, metaFile );
}


void WinMetaFile::TextOut( int x, int y, char* text, WORD count )
{
   DoStdRecord( 6 + (count+1)/2, META_TEXTOUT );
   fwrite( &count, sizeof count, 1, metaFile );
   fwrite( text, count, 1, metaFile );
   if ( count % 2 != 0 )
        fwrite( " ", 1, 1, metaFile );
   fwrite( &y, sizeof y, 1, metaFile );
   fwrite( &x, sizeof x, 1, metaFile );
}

/////////////////////////////////////////////////////////////////////////////
//                           WinMetaPlot Class

WinMetaPlot::WinMetaPlot( char* fileName, WORD Inch )
 : WinMetaFile( fileName, Inch )
{
   POINT width = { 0, 0 };

   SetMapMode( MM_ANISOTROPIC );
   SelectObject( CreatePen( PS_SOLID, width, RGB( 0, 0, 0 ) ) );
   SetBkColor( RGB( 0xFF, 0xFF, 0xFF ) );
   SetBkMode( OPAQUE );
   SetTextColor( RGB( 0, 0, 0 ) );
}


void WinMetaPlot::Axes( RECT2 & axes, RECT2 & ticks )
{
   MoveTo( axes.x1, axes.y1 );   LineTo( axes.x2, axes.y1 );
   for ( int x = axes.x1;  x <= axes.x2;  x += ticks.x1 )
       { MoveTo( x, axes.y1 - 10 );  LineTo( x, axes.y1 + 10 ); }

   MoveTo( axes.x1, axes.y1 );  LineTo( axes.x1, axes.y2 );
   for ( int y = axes.y1;  y <= axes.y2;  y += ticks.y1 )
       { MoveTo( axes.x1 - 10, y );   LineTo( axes.x1 + 10, y ); }
}

void WinMetaPlot::AxisLabels( RECT2 & axes, POINT labels, POINT scale, char* xLabel, char* yLabel )
{
   char text[15];
   int x, y;

   SetTextAlign( TA_CENTER | TA_TOP );
   for ( x = axes.x1;  x <= axes.x2;  x += labels.x )
   {
      gcvt( (float)x / scale.x, 3, text );
      TextOut( x, axes.y1 - 10, text, strlen(text) );
   }
   TextOut( (axes.x1 + axes.x2) / 2,  axes.y1 - 60, xLabel, strlen(xLabel) );

   SetTextAlign( TA_RIGHT | TA_BASELINE );
   for ( y = axes.y1;  y <= axes.y2;  y += labels.y )
   {
      gcvt( (float)y / scale.y, 3, text );
      TextOut( axes.x1 - 10, y-10, text, strlen(text) );
   }
   SetTextAlign( TA_CENTER | TA_BOTTOM );
   TextOut( axes.x1, axes.y2 + 30, yLabel, strlen(yLabel) );
}

void WinMetaPlot::LinePlot( POINT ** curves, uint * counts, uint count,
                   RECT2 & bounds, RECT2 & axes, RECT2 & ticks,
                   POINT labels, POINT scale, char* xLabel, char* yLabel )
{
   SetWindowOrg( bounds.x1, bounds.y2 );
   SetWindowExt( bounds.x2 - bounds.x1,  bounds.y1 - bounds.y2 );
   Axes( axes, ticks );
   AxisLabels( axes, labels, scale, xLabel, yLabel );
   PolyPolyline( curves, counts, count );
}


void WinMetaPlot::LinePlotIndirect( XYplot & p )
{
   SetWindowOrg( p.bounds.x1, p.bounds.y2 );
   SetWindowExt( p.bounds.x2 - p.bounds.x1,  p.bounds.y1 - p.bounds.y2 );
   Axes( p.axes, p.ticks );
   AxisLabels( p.axes, p.labels, p.scale, p.xLabel, p.yLabel );
   PolyPolyline( p.curves, p.counts, p.count );
}

