//
//  metafile.h
//
//  Header file for Windows metafile creation classes
//
//  Copyright (c) 1995 by JERICO Systems - All rights reserved
//

#ifndef  __METAFILE_H
#define  __METAFILE_H

#ifndef  __STDIO_H
#include <stdio.h>
#endif

#ifndef  __WINDOWS_H

// Substitute definitions for those in WINDOWS.H

typedef unsigned char       BYTE;
typedef unsigned short      WORD;
typedef unsigned long       DWORD;

typedef unsigned int        UINT;

typedef struct tagRECT
{
    int left;
    int top;
    int right;
    int bottom;
} RECT;

typedef struct tagPOINT
{
    int x;
    int y;
} POINT;

typedef DWORD COLORREF;

#define RGB(r,g,b) ((COLORREF)(((BYTE)(r)|((WORD)(g)<<8))|(((DWORD)(BYTE)(b))<<16)))

typedef struct tagLOGPEN
{
    UINT    lopnStyle;
    POINT   lopnWidth;
    COLORREF lopnColor;
} LOGPEN;

// Pen Styles
#define PS_SOLID            0
#define PS_DASH             1
#define PS_DOT              2
#define PS_DASHDOT          3
#define PS_DASHDOTDOT       4
#define PS_NULL             5
#define PS_INSIDEFRAME      6

typedef struct tagLOGBRUSH
{
    UINT     lbStyle;
    COLORREF lbColor;
    int      lbHatch;
} LOGBRUSH;

// Brush Styles
#define BS_SOLID            0
#define BS_NULL             1
#define BS_HOLLOW           BS_NULL
#define BS_HATCHED          2
#define BS_PATTERN          3
#define BS_INDEXED          4
#define BS_DIBPATTERN       5

// Hatch Styles
#define HS_HORIZONTAL       0
#define HS_VERTICAL         1
#define HS_FDIAGONAL        2
#define HS_BDIAGONAL        3
#define HS_CROSS            4
#define HS_DIAGCROSS        5

#define LF_FACESIZE         32
typedef struct tagLOGFONT
{
    int     lfHeight;
    int     lfWidth;
    int     lfEscapement;
    int     lfOrientation;
    int     lfWeight;
    BYTE    lfItalic;
    BYTE    lfUnderline;
    BYTE    lfStrikeOut;
    BYTE    lfCharSet;
    BYTE    lfOutPrecision;
    BYTE    lfClipPrecision;
    BYTE    lfQuality;
    BYTE    lfPitchAndFamily;
    char    lfFaceName[LF_FACESIZE];
} LOGFONT;

// Font Weight values
#define FW_DONTCARE         0
#define FW_THIN             100
#define FW_EXTRALIGHT       200
#define FW_LIGHT            300
#define FW_NORMAL           400
#define FW_MEDIUM           500
#define FW_SEMIBOLD         600
#define FW_BOLD             700
#define FW_EXTRABOLD        800
#define FW_HEAVY            900

#define FW_ULTRALIGHT       FW_EXTRALIGHT
#define FW_REGULAR          FW_NORMAL
#define FW_DEMIBOLD         FW_SEMIBOLD
#define FW_ULTRABOLD        FW_EXTRABOLD
#define FW_BLACK            FW_HEAVY

// CharSet values
#define ANSI_CHARSET        0
#define DEFAULT_CHARSET     1
#define SYMBOL_CHARSET      2
#define SHIFTJIS_CHARSET    128
#define HANGEUL_CHARSET     129
#define CHINESEBIG5_CHARSET 136
#define OEM_CHARSET         255

// OutPrecision values
#define OUT_DEFAULT_PRECIS      0
#define OUT_STRING_PRECIS       1
#define OUT_CHARACTER_PRECIS    2
#define OUT_STROKE_PRECIS       3
#if (WINVER >= 0x030a)
#define OUT_TT_PRECIS           4
#define OUT_DEVICE_PRECIS       5
#define OUT_RASTER_PRECIS       6
#define OUT_TT_ONLY_PRECIS      7
#endif  // WINVER >= 0x030a

// ClipPrecision values
#define CLIP_DEFAULT_PRECIS     0x00
#define CLIP_CHARACTER_PRECIS   0x01
#define CLIP_STROKE_PRECIS      0x02
#define CLIP_MASK               0x0F
#if (WINVER >= 0x030a)
#define CLIP_LH_ANGLES          0x10
#define CLIP_TT_ALWAYS          0x20
#define CLIP_EMBEDDED           0x80
#endif  // WINVER >= 0x030a

// Quality values
#define DEFAULT_QUALITY     0
#define DRAFT_QUALITY       1
#define PROOF_QUALITY       2

// PitchAndFamily pitch values (low 4 bits)
#define DEFAULT_PITCH       0x00
#define FIXED_PITCH         0x01
#define VARIABLE_PITCH      0x02

// PitchAndFamily family values (high 4 bits)
#define FF_DONTCARE         0x00
#define FF_ROMAN            0x10
#define FF_SWISS            0x20
#define FF_MODERN           0x30
#define FF_SCRIPT           0x40
#define FF_DECORATIVE       0x50

// Map modes
#define MM_TEXT             1
#define MM_LOMETRIC         2
#define MM_HIMETRIC         3
#define MM_LOENGLISH        4
#define MM_HIENGLISH        5
#define MM_TWIPS            6
#define MM_ISOTROPIC        7
#define MM_ANISOTROPIC      8

// Text Alignment Options
#define TA_NOUPDATECP                0x0000
#define TA_UPDATECP                  0x0001
#define TA_LEFT                      0x0000
#define TA_RIGHT                     0x0002
#define TA_CENTER                    0x0006
#define TA_TOP                       0x0000
#define TA_BOTTOM                    0x0008
#define TA_BASELINE                  0x0018

// Background Modes
#define TRANSPARENT     1
#define OPAQUE          2

// ExtTextOut options
#define ETO_GRAYED      0x0001
#define ETO_OPAQUE      0x0002
#define ETO_CLIPPED     0x0004

#endif

typedef unsigned int uint;  // convenient abbreviation

typedef struct { int x1; int y1; int x2; int y2;} RECT2;

/////////////////////////////////////////////////////////////////////////////
//                           WinMetaFile Class

typedef struct {
        DWORD key;     // 0x9AC6CDD7
        WORD  unused;  // Must be zero
        RECT2 box;     // Enclosing rectangle in LOGICAL units
        WORD  inch;     // Logical units per inch
        DWORD reserved; // Must be zero
        WORD  checkSum; // XOR sum of previous words
        } mfPreHeader;

typedef struct {
        WORD  type;
        WORD  headerSize;
        WORD  version;
        DWORD fileSize;
        WORD  maxObjectCount;
        DWORD maxRecord;
        WORD  unused;
        } mfHeader;


class WinMetaFile
{
public:
   WinMetaFile( char* fileName, WORD Inch = 100 );  // CHANGE DEFAULT ??
   ~WinMetaFile();

   WORD CreateBrushIndirect( LOGBRUSH & brush );
   WORD CreateFontIndirect( LOGFONT & font );
   WORD CreatePen( WORD style, POINT width, COLORREF color );
   WORD CreatePenIndirect( LOGPEN & pen );
   WORD CreateSolidBrush( COLORREF color );

   void Arc( int x1, int y1, int x2, int y2,
             int xbeg, int ybeg, int xend, int yend );
   void Chord( int x1, int y1, int x2, int y2,
               int xbeg, int ybeg, int xend, int yend );
   void Ellipse( int x1, int y1, int x2, int y2 );
   void ExtFloodFill( int x, int y, COLORREF color, UINT type );
   void ExtTextOut( int x, int y, UINT option, RECT & rect,
                    char* text, UINT count, int * Dx = NULL );
   void ExtTextOut( int x, int y, char* text, UINT count, int * Dx = NULL )
        { RECT r/* = {0,0,0,0}*/;  ExtTextOut( x, y, 0, r, text, count, Dx ); }
   void FloodFill( int x, int y, COLORREF color );
   void LineTo( int x, int y );
   void MoveTo( int x, int y );
   void Pie( int x1, int y1, int x2, int y2,
             int xbeg, int ybeg, int xend, int yend );
   void Polygon( const POINT * vertices, int count );
   void Polyline( const POINT * vertices, int count );
   void PolyPolygon( const POINT * vertices, int * counts, int count );
   void PolyPolyline( POINT ** curves, uint * counts, uint count );
   void Rectangle( int x1, int y1, int x2, int y2 );
   void RoundRect( int x1, int y1, int x2, int y2, int ellWide, int ellHigh );
   void SelectObject( WORD index );
   void SetBkColor( COLORREF color );
   void SetBkMode( int mode );
   void SetMapMode( WORD mode );
   void SetROP2( int mode );
   void SetTextAlign( UINT align );
   void SetTextJustification( int space, int breaks );
   void SetTextColor( COLORREF color );
   void SetWindowExt( int x, int y );
   void SetWindowOrg( int x, int y );
   void TextOut( int x, int y, char* text, WORD count );

private:
   WORD nextIndex;
   FILE * metaFile;
   mfPreHeader preHeader;
   mfHeader header;

   WinMetaFile();  // prevent default construction
   void DoStdRecord( DWORD size, WORD code, WORD newObjects = 0 );
   void UpdatePreHeader( int x1, int y1, int x2, int y2 );
   void UpdateHeader( DWORD recSize, WORD newObjects );
};

/////////////////////////////////////////////////////////////////////////////
//                           WinMetaPlot Class

typedef struct {
        RECT2 bounds;     // Overall window in logical units
        POINT** curves;   // Array of POINT arrays
        uint* counts;     // Array of curve POINT counts
        uint  count;      // Number of curves
        RECT2 axes;       // Data range window in logical units
        RECT2 ticks;      // Tick spacing in logical units
        POINT labels;     // Tick Label spacing, logical units
        POINT scale;      // physical * scale = logical
        char* xLabel;     // x axis label
        char* yLabel;     // y axis label
        } XYplot;


class WinMetaPlot : public WinMetaFile
{
public:
   WinMetaPlot( char* fileName, WORD Inch = 100 );
//   ~WinMetaPlot();

   void Axes( RECT2 & axes, RECT2 & ticks );
   void AxisLabels( RECT2 & axes, POINT labels, POINT scale, char* xLabel, char* yLabel );
   void LinePlot( POINT ** curves, uint * counts, uint count,
                  RECT2 & bounds, RECT2 & axes, RECT2 & ticks,
                  POINT labels, POINT scale, char* xLabel, char* yLabel );
   void LinePlotIndirect( XYplot & p );
};


#endif  // __METAFILE_H