//
//  metatest.cpp
//
//  Program to test and illustrate use of metafile.cpp
//
//  Copyright (c) 1995 by JERICO Systems - All rights reserved
//

#include <mem.h>
#include <math.h>
#include <string.h>

// this inclusion brings MS Windows type and data definitions
#include "metafile.h"

const COLORREF clrBlack = RGB( 0x00, 0x00, 0x00 ),
               clrBlue = RGB( 0x00, 0x00, 0xFF ),
               clrLtGreen = RGB( 0x00, 0xFF, 0x00 ),
               clrLtMagenta = RGB( 0xFF, 0x00, 0xFF ),
               clrWhite = RGB( 0xFF, 0xFF, 0xFF );

//#define CR_INDIRECT
//#define ONE_BY_ONE

void metatest()
{
   WinMetaFile mf( "test.wmf" );   // Create metafile object that
                                   // writes to disk file TEST.WMF

   //----- These next setup actions are required for placeable metafiles

   mf.SetMapMode( MM_ANISOTROPIC );  // Allows resizing of drawing
   mf.SetWindowExt( 605, 405 );      // Size of drawing area
   mf.SetWindowOrg( -1, -1 );        // Position of its __ corner


   //----- Create Pen(s), Brush(es), and Font(s) we will use

   // Note: mf.CreateXxxxYyyy(...) returns an index to the new object

   // Light Green Pen
   #ifdef CR_INDIRECT
   LOGPEN pen = { 0, {0,0}, clrLtGreen };
   mf.SelectObject( mf.CreatePenIndirect( pen ) );
   #else
   POINT wid = { 0, 0 };
   mf.SelectObject( mf.CreatePen( 0, wid, clrLtGreen ) );
   #endif

   // Light Magenta Brush
   #ifdef CR_INDIRECT
   LOGBRUSH brush = { BS_HATCHED, clrLtMagenta, HS_DIAGCROSS };
   mf.SelectObject( mf.CreateBrushIndirect( brush ) );
   #else
   mf.SelectObject( mf.CreateSolidBrush( clrLtMagenta ) );
   #endif


   // Bold Times New Roman Font
   LOGFONT font;
   memset( &font, 0, sizeof font );
   font.lfHeight = 100/6;    // 6 lines per inch
   font.lfWeight = FW_BOLD;
   font.lfQuality = PROOF_QUALITY;
   strcpy( font.lfFaceName, "Times New Roman" );
   mf.SelectObject( mf.CreateFontIndirect( font ) );


   //----- Now we are ready to draw something

   // First, a rectangle to contain everything ( 100 units per inch )
   // Bordered by the current pen, filled with the current brush
   mf.Rectangle( 0, 0, 600, 400 );

   // Then bisect it horizontally
   mf.MoveTo( 0, 200 );
   mf.LineTo( 600, 200 );

   // Next, add some aligned text
   mf.SetBkColor( clrLtMagenta );
   mf.SetBkMode( OPAQUE );
   mf.SetTextColor( clrWhite );
   char msg1[] = "Hello World!";
   mf.TextOut( 10, 10, msg1, strlen(msg1) );
   mf.SetTextAlign( TA_RIGHT | TA_BOTTOM );
   mf.TextOut( 600-10, 400-10, msg1, strlen(msg1) );

   // Now, add two polygons
   POINT polys[] = { {590,10},  {590,50},  {550,10},  {590,10},
                     {10,390},  {10,350},  {50,390},  {10,390} };
   #ifdef ONE_BY_ONE
   mf.Polygon( &polys[0], 3 );
   mf.Polygon( &polys[4], 3 );
   #else
   int counts[] = { 4, 4 };
   mf.PolyPolygon( polys, counts, 2 );
   #endif

   // Next, add some standard shapes
   mf.RoundRect( 50, 50, 250, 150, 25, 25 );
   mf.Ellipse( 350, 250, 550, 350 );
   mf.Pie( 50, 250, 250, 350, 50, 300, 150, 350 );
   mf.Arc( 50, 250, 250, 350, 150, 350, 250, 300 );
   mf.Chord( 50, 250, 250, 350, 250, 300, 150, 250 );
   mf.Arc( 50, 250, 250, 350, 150, 250, 50, 300 );

   // Finally, add some more aligned text
   mf.SetTextAlign( TA_CENTER | TA_BASELINE );
   char msg2[] = "Round Rectangle";
   mf.TextOut( 150, 100, msg2, strlen(msg2) );
   char msg3[] = "Ellipse";
   mf.TextOut( 450, 300, msg3, strlen(msg3) );
   char msg4[] = "This will be clipped!";
   RECT rect = { 400, 90, 500, 110 };  // Clipping rectangle
   mf.ExtTextOut( 450, 100, ETO_CLIPPED | ETO_OPAQUE, rect, msg4, strlen(msg4) );

   // Disk metafile will be closed automatically when mf is freed.
}


//----- Define Data Structure to Store Curves

POINT curveA[25],
      curveB[25];

POINT * curves[] = { curveA, curveB };

uint counts[2];


// Note:  Choose axes and scale so that:

//        axes.x1  <= x * scale.x  <=  axes.x2
//        axes.y1  <= y * scale.y  <=  axes.y2

// Choose bounds to allow sufficient border around axes


#ifdef CR_INDIRECT
XYplot plot = {
       { -100, -100, 1100, 1100 },       // bounds = axes + border
       curves, counts, 2,
       { 0, 0, 1000, 1000 },             // axes
       { 50, 100, 0, 0 },                // ticks
       { 100, 100 },                     // labels
       { 500, 100 },                     // scale
       "- x -",                          // xLabel
       "- y -"                           // yLabel
       };
#else
RECT2 bounds = { -100, -100, 1100, 1100 };   // axes + border
RECT2 axes   = { 0, 0, 1000, 1000 };
RECT2 ticks  = { 50, 100, 0, 0 };
POINT labels = { 100, 100 },
      scale  = { 500, 100 };
#endif


void plottest()
{
   //----- Generate and Store the curves we wish to plot

   double x, y;  // The problem space coordinates
   counts[0] = 0;   counts[1] = 0;
   for ( x = 0.0;  x <= 2.05;  x += 0.1 )  // stop at x = 2.0
   {
       int xPlot, yPlot;  // The scaled Windows graphics coordinates

       // scale x to chosen window units
#ifdef CR_INDIRECT
       xPlot = (int)floor( x * plot.scale.x + 0.5 );
#else
       xPlot = (int)floor( x * scale.x + 0.5 );
#endif

       // compute and store curve A point
       y = x*x;
#ifdef CR_INDIRECT
       yPlot = (int)floor( y * plot.scale.y + 0.5 );  // scale y
#else
       yPlot = (int)floor( y * scale.y + 0.5 );
#endif
       curveA[counts[0]].x = xPlot;
       curveA[counts[0]].y = yPlot;
       ++counts[0];

       // compute and store curve B point
       y = x*x*x;
#ifdef CR_INDIRECT
       yPlot = (int)floor( y * plot.scale.y + 0.5 );      // scale y
#else
       yPlot = (int)floor( y * scale.y + 0.5 );
#endif
       curveB[counts[1]].x = xPlot;
       curveB[counts[1]].y = yPlot;
       ++counts[1];
   }

   //----- Plot the Curves and Annotate

   WinMetaPlot mp( "testplot.wmf" );     // Create a WinMetaPlot object that
                                         // writes to disk file TESTPLOT.WMF
   // First, plot the curves
   // Note SetMapMode(), SetWindowOrg(), SetWindowExt() unnecessary here
#ifdef CR_INDIRECT
   mp.LinePlotIndirect( plot );
#else
   mp.LinePlot( curves, counts, 2, bounds, axes, ticks, labels, scale,
                 "- x -", "- y -" );
#endif

   // Then, add some extra labels
   char lab1[] = "y = x*x";
   mp.SetTextAlign( TA_RIGHT | TA_TOP );
   mp.TextOut( 950, 200, lab1, strlen(lab1) );
   char lab2[] = "y = x*x*x";
   mp.SetTextAlign( TA_RIGHT | TA_BOTTOM );
   mp.TextOut( 950, 800, lab2, strlen(lab2) );

   // Disk metafile will be closed automatically when mp is freed.
}


void main()
{
   metatest();
   plottest();
}